pro make_linearity_tables, Event
;
;   uses images from a series of exposure times to
;   compute linearity tables
;
;   DPC  20041003
;        20041030 updated with October 2004 image size
;        20041031 updated with softer saturation limit (30% max slope instead of 60%)
;        20050105 added bias frame subtraction to flatten top/bottom ski slope problem
;        20050328 replaced bias frame subtraction with *addition* of difference between mean
;                 (flat-bias) for RAW1 (first read) to realize absolute fluxes
;        20050614 modified to read mean BIAS RAW1 frame, but then each mean RAW1 frame
;                   for every exposure time
;        20050615 pushed saturation higher with 15% slope test in place of 30%
;        20050713 added ability to dump all data, fits for one pixel for testing
;                 increased number of terms in fit and added weighting by RN and photons
;        20050714 fixed weighting problem in cc fit, modified saturation detection, reset to 30%
;        20051124 modified to use safe_math where ever image calculations are done, and
;                   to make saturation values deleted data if input that way, also added test for
;                   negative saturation value - set to deleted data if found
;		 20071217 imported from make_linearity_tables.pro into MSP_BDP widget program
;		 20080523 modified to use set of 4 input data array pointers
;
;		 20100709 added ramp cleaning
;
;					note: data cube pointed to on entry is already (RR-D2) type (ie corrected for D1, B1)
;
;   tables will be in the form of images with 2N+1 layers, where N is the fit order:
;     saturation - contains count limit of linearity for each pixel
;     cc2 - curvature coefficient for regression of counts vs time
;     cc3
;     cc4
;     tc0 - offset coefficient for regression of time vs counts
;     tc1 - slope coefficient for regression of time vs counts
;     tc2 - curvature coefficient for regression of time vs counts
;     tc3
;     tc4
;
;   to calculate the count correction needed in a given pixel
;   in an image, the following equation applies:
;
;   t = sum[t_i * counts^i] ; i=0, N
;   additive_correction = -{sum[c_i * t^i] ; i=2, N}
;
;   for pixels flagged in the bad pixel mask, all layers will contain
;   the value 0.0
;
;----------------------
;MSP_BDP_print_text, Event, "Make Linearity Tables Requested"
;
WIDGET_CONTROL, Event.top, GET_UVALUE = PTR_struct
n_ptr = PTR_struct.PTR_night
s_ptr = PTR_struct.PTR_summary
l_ptr = PTR_struct.PTR_Linearity
r_ptr = PTR_struct.PTR_RAW1
;
d_ptr_0 = (*l_ptr).lin_data_ptr_0		; pointers to the structures that holds the big data 3D array
d_ptr_1 = (*l_ptr).lin_data_ptr_1
d_ptr_2 = (*l_ptr).lin_data_ptr_2
d_ptr_3 = (*l_ptr).lin_data_ptr_3
;
;	recover header template
;
n_lin_header = (*d_ptr_0).n_header
lin_header = (*d_ptr_0).lin_header[0:n_lin_header-1]
;
;	set up LED progress bar
;
P6_LED_sensitize, Event, 2, 1
LED_wid = (*s_ptr).P6_LED_wid[2]
colors = strarr(10)
colors[*] = 'yellow'
WCT_LED_stack, (*s_ptr).P6_LED_wid[2], colors
;
LED_scale = float(1025)/10.0
;
;   number of the order of the fits
;
norder = 4
;
;   use weighting if nweight = 1; else no weighting
;
nweight = 1
RN = 17.8          ; effective electrons per read
cg = 8.21          ; electrons per ADU
;
image_size = SIZE( (*d_ptr_0).BPM_image )
ncol = image_size[1]
nrow = image_size[2]
;
nncol = ncol
nnrow = nrow
;
;   saturation test = % of maximum slope value
;
slope_test = 0.30
;
;   deleted data value
;
deleted_data_value = -1.0e6
;
;   test mode? TBD - this would be nice to go to a file, OK to go to plot, both using pixel selected earlier
;
itest = 1
;
;   reserve space for results table
;
fit_table = fltarr(2*norder+1,ncol,nrow)
;
;   set output table files
;
outfile = STRTRIM(STRING(format='(I8.8)',(*n_ptr).night_desig),2)+'_Lin_'
;
;	attach directory
;
root_reduced = (*s_ptr).root_reduced
this_lin = (*l_ptr).this_lin
;
night_desig = (*n_ptr).night_desig
night_desig_text = STRING(format='(I8.8)',night_desig)
;
lin_directory = '6_Linearity_Tables' + PATH_SEP() + night_desig_text
subdirectory = "linearity_table_images"
;
lin_path = root_reduced + lin_directory + PATH_SEP() + subdirectory
;
FILE_MKDIR, lin_path		; just in case it doesn't exist
;
outfile = lin_path + PATH_SEP() + outfile
;
nfile = (*d_ptr_0).n_images-1
time = (*d_ptr_0).lin_expt[0:nfile-1]
counts = fltarr(100)
;
;   get the data base information
;
ifile = 0
t = 0.0
;
big_string = ''
;
badpix = (*d_ptr_0).BPM_image					; long
biaspix = (*d_ptr_0).bias_raw1_image			; double
;
;	data file
;
; (*d_ptr).lin_image[ifile, ic, ir]
;
;   for each pixel, extract the run of (RAW2-reset) as a function of exposure time
;
vec = fltarr(nfile)
slope = fltarr(nfile)
;
ncbad = 0
ntbad = 0
nbp = 0
for icol = 0, nncol-1 do begin
    ;
    this_LED = ROUND( float(icol) / LED_scale)
    this_LED = 0 > (this_LED < 9)
    ;
    colors[this_LED] = 'orange'
    WCT_LED_stack, (*s_ptr).P6_LED_wid[2], colors
    ;
    for irow = 0, nnrow-1 do begin
    	;
    	if(irow eq ROUND(nnrow/4)) then begin
    		colors[this_LED] = 'blue'
    		WCT_LED_stack, (*s_ptr).P6_LED_wid[2], colors
    	endif
    	if(irow eq ROUND(nnrow/2)) then begin
    		colors[this_LED] = 'violet'
    		WCT_LED_stack, (*s_ptr).P6_LED_wid[2], colors
    	endif
    	if(irow eq ROUND(3*nnrow/4)) then begin
    		colors[this_LED] = 'white'
    		WCT_LED_stack, (*s_ptr).P6_LED_wid[2], colors
    	endif
        ;
        ;  load all the data for the pixel into one vector
        ;
        time = (*d_ptr_0).lin_expt[0:nfile-1]
        ;
        ;	pack vector from up to 4 data arrays
        ;
        vec = REFORM( (*d_ptr_0).lin_image_0[0:(25 < (nfile-1)),icol,irow] )    ; load the vector of value vs exp time
        if(nfile gt 26) then begin
        	vec = [vec, REFORM( (*d_ptr_1).lin_image_1[0:(24 < (nfile-27)),icol,irow] )]
        endif
        ;
        if(nfile gt 51) then begin
        	vec = [vec, REFORM( (*d_ptr_2).lin_image_2[0:(24 < (nfile-52)),icol,irow] )]
        endif
        ;
        if(nfile gt 76) then begin
        	vec = [vec, REFORM( (*d_ptr_3).lin_image_3[0:(24 < (nfile-77)),icol,irow] )]
        endif
        ;
        ;	check that at least 50% of the images are not deleted
        ;
        t_count = N_ELEMENTS(vec)
        jnd = WHERE(vec ne deleted_data_value, g_count)
        ;
        if( FLOAT(g_count) ge 0.5*FLOAT(t_count) ) then d_test = 1 else d_test = 0
        ;
        ;     if a bad pixel, clear the table constants
        ;
        if(badpix[icol,irow] or NOT(d_test)) then begin
            fit_table[0,icol,irow] = deleted_data_value
            fit_table[1:2*norder,icol,irow] = 0.0
;            fit_table[2*norder+1,icol,irow] = deleted_data_value
            nbp++
        endif else begin
            ;
            ;	get rid of deleted pixels (to save the others)
            ;
            g_ind = WHERE(vec[*] ne deleted_data_value, g_count)
            vec = vec[g_ind]
            time = time[g_ind]
            ;
            ;	20100709 - clean the ramp of deviant pixels via rolling polynomial fit
            ;
            clean_ramp, time, vec
            ;
            ;
            ;   if test mode, dump vector to save file
            ;
;            if (itest) then begin
;                if(ix_test eq icol and iy_test eq irow) then begin
;                   for i_test = 0, g_count-1 do begin
;                     printf, test_lun, FORMAT="(F10.6,1x,F12.6)",time[i_test],vec[i_test]
;                   endfor
;                endif
;            endif
            ;
            ;  determine saturation value
            ;
            for ifile = 0, g_count-2 do begin
                 slope[ifile] = (vec[ifile+1] - vec[ifile]) / (time[ifile+1] - time[ifile])
            endfor
            ;
            ;   compute max slope
            ;
            max_slope = max(slope)
            ;
            ;   here is the test. Once the slope is less than some
            ;   fraction of the max slope, and this persists for two of the next three
            ;   samples, we assume saturation has set in
            ;
            ifile_bad = 0
            st = slope_test*max_slope
            for ifile = 0, g_count-2 do begin
                 ; do the tests
                 ;
                 if(slope[ifile] ge st) then it0 = 1 else it0 = 0
                 if(slope[(ifile+1)<(g_count-1)] ge st) then it1 = 1 else it1 = 0
                 if(slope[(ifile+2)<(g_count-1)] ge st) then it2 = 1 else it2 = 0
                 if(slope[(ifile+3)<(g_count-1)] ge st) then it3 = 1 else it3 = 0
                ;
                if((it0+it1+it2+it3) ge 3) then ifile_bad = (ifile+2) < (g_count-1)
            endfor
            ;
            ;  ifile_bad is the saturation limit. The limit for fitting is two samples more
            ;
            ifile_fit = (ifile_bad + 2) < (g_count-1)
            fit_table[0,icol,irow] = vec[ifile_bad]      ; last point is this pixel's saturation limit
            ;

            ;   test for negative saturation, if found set to deleted data value
            ;
            if(fit_table[0,icol,irow] le 0.0) then begin
                fit_table[0,icol,irow] = deleted_data_value
                nbp++
            endif
            ;
            ;  do regression of the non-saturated points going both ways (time to ADUs and ADUs to time]
            ;
            ztime = time[0:ifile_fit]
            zvec = vec[0:ifile_fit]
            nnorder = norder
            ;
            ;   compute weights if turned on
            ;
            weight = fltarr(ifile_fit)
            if (nweight eq 1) then begin
                weight = sqrt((RN/cg)^2 + zvec/cg)
                if (itest and (*l_ptr).xx eq icol and (*l_ptr).yy eq irow) then begin
                    for iii = 0, ifile_fit-1 do begin
                       MSP_BDP_print_text, Event, "Weight("+STRING(format='(I5.1)',iii)+") = "+$
                       					STRING(format='(F12.6)',weight[iii])
                    endfor
                endif
            endif else begin
                weight = 1.
            endelse
            ;
            cc = poly_fit(ztime,zvec,nnorder,/DOUBLE, STATUS=cc_status,MEASURE_ERRORS=weight)
            ;
            ;	for some reason, cc is returned at 2-D array, with first index = 1, so reform
            ;
            cc = REFORM(cc)
            ;
            ;cc = poly_fit(ztime,zvec,nnorder,/DOUBLE, STATUS=cc_status)
            ;
;            if(cc_status ne 0) then begin
;            	MSP_BDP_print_text, Event, "icol, irow =("+STRING(format='(I4.1,",",I4.1)',icol,irow)+$
;            								") status of cc fit = "+STRING(format='(I4.1)',cc_status)
;            endif
            if(cc_status ne 0) then begin                 ; something really wrong, zero correction
                cc = fltarr(norder+1)
                cc[2:norder] = 0.0
                fit_table[0,icol,irow] = deleted_data_value
                ncbad++
            endif
			;
			;      this fit unweighted, since time not uncertain
			;
            ztime = time[0:ifile_fit]
            zvec = vec[0:ifile_fit]
            nnorder = norder
            tc = poly_fit(zvec,ztime,nnorder,/DOUBLE,STATUS=tc_status)
            ;
;            if(tc_status ne 0) then begin
;            	MSP_BDP_print_text, Event, "icol, irow =("+STRING(format='(I4.1,",",I4.1)',icol,irow)+$
;            								") status of cc fit = "+STRING(format='(I4.1)',tc_status)
;            endif
            if(tc_status ne 0) then begin
                tc = fltarr(norder+1)
                tc[0:norder] = 0.0
                fit_table[0,icol,irow] = deleted_data_value
                ntbad++
            endif
            ;
            ;    save key coefficients
            ;
            fit_table[1,icol,irow] = cc[2]
            case norder of
                2: begin
                   fit_table[2,icol,irow] = tc[0]
                   fit_table[3,icol,irow] = tc[1]
                   fit_table[4,icol,irow] = tc[2]
                   end
                3: begin
                   fit_table[2,icol,irow] = cc[3]
                   fit_table[3,icol,irow] = tc[0]
                   fit_table[4,icol,irow] = tc[1]
                   fit_table[5,icol,irow] = tc[2]
                   fit_table[6,icol,irow] = tc[3]
                   end
                4: begin
                   fit_table[2,icol,irow] = cc[3]
                   fit_table[3,icol,irow] = cc[4]
                   fit_table[4,icol,irow] = tc[0]
                   fit_table[5,icol,irow] = tc[1]
                   fit_table[6,icol,irow] = tc[2]
                   fit_table[7,icol,irow] = tc[3]
                   fit_table[8,icol,irow] = tc[4]
                   end
             endcase
            ;
            ;	do the readout time calculation
            ;
;            if(cc_status eq 0) then begin
;            	complex_roots = FZ_ROOTS(cc)
;            	ind_1 = WHERE(IMAGINARY(complex_roots) eq 0.0,count)
;            	if(count ge 1) then begin
;            		rea_roots = DOUBLE(complex_roots[ind_1])
;            		jnd_2 = WHERE(rea_roots lt 0.0,ncount)
;            		if(ncount ge 1) then begin
;            			neg_rea_roots = ABS(rea_roots[jnd_2])
;            			fit_table[9,icol,irow] = MIN(neg_rea_roots)	; negative real root closest to zero
;            		endif else begin
;            			fit_table[9,icol,irow] = deleted_data_value
;            		endelse
;            	endif else begin
;            		fit_table[9,icol,irow] = deleted_data_value
;            	endelse
;            endif else fit_table[9,icol,irow] = deleted_data_value
            ;
            ;   if test mode, add these to the bottom of the output file
            ;
            if (itest) then begin
              if((*l_ptr).xx eq icol and (*l_ptr).yy eq irow) then begin
               MSP_BDP_print_text, Event, "SAT   = "+STRING(format='(F12.2)',fit_table[0,icol,irow])
               MSP_BDP_print_text, Event, "CC[0] = "+STRING(format='(F11.5)',cc[0])
               MSP_BDP_print_text, Event, "CC[1] = "+STRING(format='(F12.4)',cc[1])
               MSP_BDP_print_text, Event, "CC[2] = "+STRING(format='(F11.5)',cc[2])
               if (norder ge 3) then MSP_BDP_print_text, Event, "CC[3] = "+STRING(format='(F11.6)',cc[3])
               if (norder ge 4) then MSP_BDP_print_text, Event, "CC[4] = "+STRING(format='(F11.7)',cc[4])
               MSP_BDP_print_text, Event, "TC[0] = "+STRING(format='(F12.8)',tc[0])
               MSP_BDP_print_text, Event, "TC[1] = "+STRING(format='(E15.7)',tc[1])
               MSP_BDP_print_text, Event, "TC[2] = "+STRING(format='(E15.7)',tc[2])
               if (norder ge 3) then MSP_BDP_print_text, Event, "TC[3] = "+STRING(format='(E15.7)',tc[3])
               if (norder ge 4) then MSP_BDP_print_text, Event, "TC[4] = "+STRING(format='(E15.7)',tc[4])
;			   MSP_BDP_print_text, Event, "Read Time = "+STRING(format='(F12.5)',fit_table[9,icol,irow])
              endif
            endif
            if(icol eq (*l_ptr).xx and irow eq (*l_ptr).yy) then begin
            	MSP_BDP_print_text, Event, "Saturation limit for ("+STRING(format='(I4.1)',(*l_ptr).xx)+$
            				','+STRING(format='(I4.1)',(*l_ptr).yy)+') = '+$
            				STRING(format='(F12.4)',vec[ifile_bad])+" at time = "+$
            				STRING(format='(F12.5)',time[ifile_bad])
            	;
            	;	and plot it
            	;
            	device, decomposed=0
				color_red = REFORM((*r_ptr).color_rgb[0,*])
				color_green = REFORM((*r_ptr).color_rgb[1,*])
				color_blue = REFORM((*r_ptr).color_rgb[2,*])
				;
				tvlct, color_red, color_green, color_blue, 0
				;
            	win_wid = (*s_ptr).P6_draw_window_id
				WSET, win_wid
				;
				ysat = vec[ifile_bad]
            	xp = [time[ifile_bad],time[ifile_bad]]
            	yp = [0.,vec[ifile_bad]]
            	PlotS, xp, yp, color=2, linestyle=2, thick=3.0
            	xp = [-1.,time[ifile_bad]]
            	yp = [vec[ifile_bad],vec[ifile_bad]]
				PlotS, xp, yp, color=5, linestyle=2, thick=3.0
				;
				xx = time[ifile_bad]
				yy = vec[ifile_bad]+400
				XYoutS, xx, yy, "S", color=2
				;
				;	plot read time
				;
;				rd_time = -1.0 * fit_table[9,icol, irow]
;				xp = [rd_time, rd_time]
;				yp = [0., 0.2*ysat]
;				PlotS, xp, yp, color=6, linestyle=2, thick=3.0
;				xx = rd_time
;				yy = 0.225 * ysat
;				XYoutS, xx, yy, "Trd", color=6
            endif
            ;
        endelse
    endfor
    ;
    colors[this_LED] = 'green'
    WCT_LED_stack, (*s_ptr).P6_LED_wid[2], colors
endfor
;
;   how best to save this table?
;
;   for now, do as 2*n+1 image files
;
slice = ''
outdata = fltarr(nncol,nnrow)
for i = 0, 2*norder do begin
	;
    slice = outfile + string(format="(i1)",i) + ".fits"
    ;
    ;	register the linearity table in the list
    ;
    if(i eq 0) then begin
    	(*l_ptr).lin_table_0_path[ (*l_ptr).n_lin_tables ] = slice
    	(*l_ptr).n_lin_tables++
    endif
    ;
    outdata = fit_table[i,*,*]
    outdata = reform(outdata)     ; reduces dimensionality to 2-D for saving to disk
    ;
    ;   modify OBJECT and FILENAME in FITS header
    ;
    name_o = 'OBJECT'
    name_f = "FILENAME"
    f_value = FILE_BASENAME(slice)
    f_comment = 'Written by MAKE_LINEARITY_TABLES.PRO'
    fxaddpar, lin_header, name_f, f_value, f_comment
    ;
    case i of
       0: begin
         o_value = 'saturation'
         o_comment = 'Limit from MAKE_LINEARITY_TABLES.PRO'
          end
       1: begin
           o_value = 'cc[2]'
           o_comment = 'Quadratic Correction from MAKE_LINEARITY_TABLES.PRO'
          end
       2: begin
            case norder of
              2: begin
               o_value = 'tc[0]'
               o_comment = 'Time Offset from MAKE_LINEARITY_TABLES.PRO'
                 end
              3: begin
               o_value = 'cc[3]'
               o_comment = 'Cubic Correction from MAKE_LINEARITY_TABLES.PRO'
                   end
                4: begin
                    o_value = 'cc[3]'
               		o_comment = 'Cubic Correction from MAKE_LINEARITY_TABLES.PRO'
             	end
           endcase
          end
       3: begin
            case norder of
              2: begin
               o_value = 'tc[1]'
               o_comment = 'Time Slope from MAKE_LINEARITY_TABLES.PRO'
                 end
              3: begin
               o_value = 'tc[0]'
               o_comment = 'Time Offset from MAKE_LINEARITY_TABLES.PRO'
                   end
                4: begin
                    o_value = 'cc[4]'
               o_comment = 'Quartic Correction from MAKE_LINEARITY_TABLES.PRO'
             end
           endcase
          end
       4: begin
            case norder of
              2: begin
               o_value = 'tc[2]'
               o_comment = 'Time Quadratic from MAKE_LINEARITY_TABLES.PRO'
                 end
              3: begin
               o_value = 'tc[1]'
               o_comment = 'Time Slope from MAKE_LINEARITY_TABLES.PRO'
                   end
                4: begin
                    o_value = 'tc[0]'
               o_comment = 'Time Offset from MAKE_LINEARITY_TABLES.PRO'
             end
           endcase
          end
       5: begin
            case norder of
               3: begin
                 o_value = 'tc[2]'
                 o_comment = 'Time Quadratic from MAKE_LINEARITY_TABLES.PRO'
                  end
               4: begin
                 o_value = 'tc[1]'
                 o_comment = 'Time Slope from MAKE_LINEARITY_TABLES.PRO'
                  end
            endcase
           end
        6: begin
            case norder of
               3: begin
                 o_value = 'tc[3]'
                 o_comment = 'Time Cubic from MAKE_LINEARITY_TABLES.PRO'
                  end
               4: begin
                 o_value = 'tc[2]'
                 o_comment = 'Time Quadratic from MAKE_LINEARITY_TABLES.PRO'
                  end
            endcase
           end
        7: begin
            o_value = 'tc[3]'
            o_comment = 'Time Cubic from MAKE_LINEARITY_TABLES.PRO'
           end
        8: begin
            o_value = 'tc[4]'
            o_comment = 'Time Quartic from MAKE_LINEARITY_TABLES.PRO'
           end
;        9: begin
;        	o_value = 'Read Time'
;        	o_comment = 'Fitted Single Read Time'
;           end
    endcase
    ;
    fxaddpar, lin_header, name_o, o_value, o_comment
    FXADDPAR, lin_header, "HISTORY", "Linearity Table created "+SYSTIME()
    ;
    writefits, slice, outdata, lin_header
endfor
;
MSP_BDP_print_text, Event, "Number of Bad pixels (no linearity correction) = "+STRING(format='(I5.1)',nbp)
MSP_BDP_print_text, Event, "Number of C-fits bad = "+STRING(format='(I5.1)',ncbad)+$
                                " Number of T-fits bad = "+STRING(format='(I5.1)',ntbad)
end


